// Generated by CoffeeScript 1.4.0
/*
Dialects are responsible for compiling an AST to a SQL string compatible with
a particular DBMS. They are rarely used directly, instead a query is usually
bound to an `engine <Engines>`_ that will delegate compiling to it's dialect
instance.
*/

var BaseDialect, Delete, Field, Insert, MySQLDialect, PostgresDialect, PrettyDialect, Relation, SQLite3Dialect, Select, Update, keywords, kwFile, read, _ref,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; };

read = require('fs').readFileSync;

kwFile = __dirname + '/sql_keywords.txt';

keywords = read(kwFile, 'ascii').split('\n').filter(Boolean);

_ref = require('./nodes'), Select = _ref.Select, Update = _ref.Update, Delete = _ref.Delete, Insert = _ref.Insert, Relation = _ref.Relation, Field = _ref.Field;

BaseDialect = (function() {
  var Visitor, doubleQuote, needsQuote;

  function BaseDialect() {}

  BaseDialect.prototype.reset = function() {};

  BaseDialect.prototype.compile = function(root) {
    var text, visitor;
    visitor = new Visitor(this);
    text = visitor.compile(root);
    return [text, visitor.params];
  };

  BaseDialect.prototype.renderString = function(s) {
    var path;
    path = this.path.map(function(p) {
      var _ref1;
      return (_ref1 = p.constructor) != null ? _ref1.name : void 0;
    }).join(' > ');
    this.path = [];
    throw new Error("raw string compiled! " + path);
  };

  needsQuote = /\s|"|\./;

  doubleQuote = /"/g;

  BaseDialect.prototype.quote = function(s) {
    if ((s != null ? s.match(needsQuote) : void 0) || this.isKeyword(s)) {
      return '"' + s.replace(doubleQuote, '\\"') + '"';
    } else {
      return s;
    }
  };

  BaseDialect.prototype.isKeyword = function(word) {
    return keywords.indexOf(word.toUpperCase()) !== -1;
  };

  BaseDialect.prototype.operator = function(op) {
    switch ((op = op.toUpperCase())) {
      case 'NE':
      case '!=':
      case '<>':
        return '!=';
      case 'EQ':
      case '=':
        return '=';
      case 'LT':
      case '<':
        return '<';
      case 'GT':
      case '>':
        return '>';
      case 'LTE':
      case '<=':
        return '<=';
      case 'GTE':
      case '>=':
        return '>=';
      case 'LIKE':
      case 'ILIKE':
      case 'IN':
      case 'NOT IN':
      case 'IS':
      case 'IS NOT':
        return op;
      default:
        throw new Error("Unsupported comparison operator: " + op);
    }
  };

  BaseDialect.prototype.placeholder = function(position) {
    return "$" + position;
  };

  Visitor = (function() {

    function Visitor(dialect) {
      this.dialect = dialect;
      this.path = [];
      this.params = [];
    }

    Visitor.prototype.compile = function(node, allowOverride) {
      var custom, name, string, _ref1, _ref2;
      if (allowOverride == null) {
        allowOverride = true;
      }
      this.path.push(node);
      name = node != null ? (_ref1 = node.__proto__) != null ? (_ref2 = _ref1.constructor) != null ? _ref2.name : void 0 : void 0 : void 0;
      if (allowOverride && name && (custom = this.dialect['render' + name])) {
        string = custom.call(this, node);
      } else {
        string = node.compile(this, this.path);
      }
      this.path.pop(node);
      return string;
    };

    Visitor.prototype.maybeParens = function(it) {
      if (/\s/.exec(it)) {
        return "(" + it + ")";
      } else {
        return it;
      }
    };

    Visitor.prototype.operator = function(string) {
      return this.dialect.operator(string);
    };

    Visitor.prototype.parameter = function(val) {
      this.params.push(val);
      return this.dialect.placeholder(this.params.length);
    };

    Visitor.prototype.quote = function(string) {
      return this.dialect.quote(string, this.path);
    };

    return Visitor;

  })();

  return BaseDialect;

})();

PrettyDialect = (function(_super) {

  __extends(PrettyDialect, _super);

  function PrettyDialect() {
    return PrettyDialect.__super__.constructor.apply(this, arguments);
  }

  PrettyDialect.prototype.renderJoin = function(node) {
    return "\n" + this.compile(node, false);
  };

  PrettyDialect.prototype.renderWhere = function(node) {
    return "\n" + this.compile(node, false);
  };

  PrettyDialect.prototype.renderHaving = function(node) {
    return "\n" + this.compile(node, false);
  };

  PrettyDialect.prototype.renderOrderBy = function(node) {
    return "\n" + this.compile(node, false);
  };

  PrettyDialect.prototype.renderGroupBy = function(node) {
    return "\n" + this.compile(node, false);
  };

  PrettyDialect.prototype.renderRelationSet = function(node) {
    return "\n" + this.compile(node, false);
  };

  PrettyDialect.prototype.renderSelectColumnSet = function(node) {
    var glue, last, lines, size, text, thisLine, thisLineLength, _i, _len, _ref1;
    glue = node.glue;
    last = node.nodes.length;
    lines = [];
    thisLine = [];
    thisLineLength = 81;
    _ref1 = node.nodes;
    for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
      node = _ref1[_i];
      text = this.compile(node);
      size = text.length + glue.length;
      if (thisLineLength + size > 50) {
        lines.push(thisLine.join(glue));
        thisLine = [];
        thisLineLength = 0;
      }
      thisLineLength += size;
      thisLine.push(text);
    }
    lines.shift();
    lines.push(thisLine.join(glue));
    return lines.join("\n  ");
  };

  return PrettyDialect;

})(BaseDialect);

PostgresDialect = (function(_super) {

  __extends(PostgresDialect, _super);

  function PostgresDialect() {
    return PostgresDialect.__super__.constructor.apply(this, arguments);
  }

  PostgresDialect.prototype.operator = function(op) {
    switch (op.toLowerCase()) {
      case 'hasKey':
        return '?';
      case '->':
        return '->';
      default:
        return PostgresDialect.__super__.operator.call(this, op);
    }
  };

  PostgresDialect.prototype.isKeyword = function(s) {
    return (s != null) && s !== '*';
  };

  return PostgresDialect;

})(BaseDialect);

MySQLDialect = (function(_super) {

  __extends(MySQLDialect, _super);

  function MySQLDialect() {
    return MySQLDialect.__super__.constructor.apply(this, arguments);
  }

  MySQLDialect.prototype.placeholder = function() {
    return '?';
  };

  MySQLDialect.prototype.quote = function(s, path) {
    /*
        MySQL has two special cases for quoting:
         - The column names in an insert column list are not quoted
         - table and field names are quoted with backticks.
    */

    var node;
    node = path[path.length - 1];
    if (s === '*' || path.some(function(node) {
      return node instanceof Insert.ColumnList;
    })) {
      return s;
    } else if (node instanceof Field || node instanceof Relation) {
      return "`" + s + "`";
    } else {
      return MySQLDialect.__super__.quote.apply(this, arguments);
    }
  };

  return MySQLDialect;

})(BaseDialect);

SQLite3Dialect = (function(_super) {

  __extends(SQLite3Dialect, _super);

  function SQLite3Dialect() {
    return SQLite3Dialect.__super__.constructor.apply(this, arguments);
  }

  SQLite3Dialect.prototype.placeholder = function() {
    return '?';
  };

  SQLite3Dialect.prototype.renderInsertData = function(node) {
    var string;
    if (node.nodes.length < 2) {
      return node.compile(this, this.path);
    } else {
      node.glue = ' UNION ALL SELECT ';
      string = node.compile(this, this.path).replace('VALUES', 'SELECT').replace(/[()]/g, '');
      node.glue = ', ';
      return string;
    }
  };

  return SQLite3Dialect;

})(BaseDialect);

module.exports = {
  base: BaseDialect,
  pretty: PrettyDialect,
  postgres: PostgresDialect,
  mysql: MySQLDialect,
  sqlite3: SQLite3Dialect
};
